@@ -63,7 +63,11 @@ impl std::fmt::Display for Error {
                 expected, at, found
             ),
             Error::Unsupported(ref ty) => {
-                write!(f, "the type '{}' is not supported by the library", ty)
+                write!(
+                    f,
+                    "decoding of type '{}' is not supported by the library",
+                    ty
+                )
             }
             Error::Deserialize(ref e) => e.fmt(f),
             Error::Utf8(ref e) => e.fmt(f),
@@ -426,11 +430,11 @@ impl<'a, 'de> Deserializer<'de> for &'a mut Decoder<'de> {
         visitor.visit_byte_buf(self.decode_bytes()?.to_vec())
     }
 
-    fn deserialize_option<V>(self, _: V) -> Result<V::Value, Self::Error>
+    fn deserialize_option<V>(self, visitor: V) -> Result<V::Value, Self::Error>
     where
         V: serde::de::Visitor<'de>,
     {
-        Err(Error::Unsupported("Option<_>"))
+        visitor.visit_some(self)
     }
 
     fn deserialize_unit<V>(self, visitor: V) -> Result<V::Value, Self::Error>
@@ -442,13 +446,25 @@ impl<'a, 'de> Deserializer<'de> for &'a mut Decoder<'de> {
 
     fn deserialize_unit_struct<V>(
         self,
-        _: &'static str,
+        name: &'static str,
         visitor: V,
     ) -> Result<V::Value, Self::Error>
     where
         V: serde::de::Visitor<'de>,
     {
-        visitor.visit_unit()
+        // Get the decoder's position before we decode the name.
+        let at = self.pos();
+
+        let found = self.decode_bytes()?;
+        if found != name.as_bytes() {
+            Err(Error::Wanted {
+                at,
+                expected: name,
+                found: String::from_utf8(found.to_vec())?,
+            })
+        } else {
+            visitor.visit_unit()
+        }
     }
 
     fn deserialize_newtype_struct<V>(
@@ -459,7 +475,7 @@ impl<'a, 'de> Deserializer<'de> for &'a mut Decoder<'de> {
     where
         V: serde::de::Visitor<'de>,
     {
-        visitor.visit_unit()
+        visitor.visit_newtype_struct(self)
     }
 
     // Sequences and tuple types are deserialized as a sequence.
@@ -614,7 +630,8 @@ impl<'a, 'de> MapAccess<'de> for MapDecoder<'a, 'de> {
                 self.de.advance(1);
                 Ok(None)
             }
-            Some(_) => seed.deserialize(&mut *self.de).map(Some),
+            Some(b'0'..=b'9') => seed.deserialize(&mut *self.de).map(Some),
+            Some(_) => Err(Error::Malformed),
             _ => Err(Error::EOF),
         }
     }
@@ -629,6 +646,8 @@ impl<'a, 'de> MapAccess<'de> for MapDecoder<'a, 'de> {
 
 #[cfg(test)]
 mod test {
+    use std::collections::HashMap;
+
     use serde::Deserialize;
 
     use super::{Decoder, Error};
@@ -706,6 +725,12 @@ mod test {
         assert_eq!(de.next(), Some(b'x'));
     }
 
+    #[test]
+    fn decode_bytes_empty() {
+        let mut de = Decoder::new(b"0:");
+        assert_eq!(de.decode_bytes(), Ok(b"".as_slice()));
+    }
+
     #[test]
     fn decode_bytes_err() {
         let mut de = Decoder::new(b"4:foo");
@@ -757,4 +782,86 @@ mod test {
 
         assert_eq!(Person::deserialize(&mut de), Ok(jerry));
     }
+
+    #[test]
+    fn deserialize_some() {
+        #[derive(Debug, PartialEq, Deserialize)]
+        struct Person {
+            name: Option<String>,
+            age: u8,
+        };
+
+        let mut de = Decoder::new(b"d4:name5:Jerry3:agei50ee");
+        assert_eq!(
+            Person::deserialize(&mut de),
+            Ok(Person { name: Some("Jerry".to_string()), age: 50 })
+        )
+    }
+
+    #[test]
+    fn deserialize_none() {
+        #[derive(Debug, PartialEq, Deserialize)]
+        struct Person {
+            name: Option<String>,
+            age: u8,
+        };
+
+        let mut de = Decoder::new(b"d3:agei50ee");
+        assert_eq!(
+            Person::deserialize(&mut de),
+            Ok(Person { name: None, age: 50 })
+        )
+    }
+
+    #[test]
+    fn deserialize_unit_struct_ok() {
+        #[derive(Debug, PartialEq, Deserialize)]
+        struct Unit;
+
+        let mut de = Decoder::new(b"4:Unit");
+        assert_eq!(Unit::deserialize(&mut de), Ok(Unit));
+    }
+
+    #[test]
+    fn deserialize_unit_struct_err() {
+        #[derive(Debug, PartialEq, Deserialize)]
+        struct Unit;
+
+        let mut de = Decoder::new(b"3:Foo");
+        assert_eq!(
+            Unit::deserialize(&mut de),
+            Err(Error::Wanted {
+                at: 0,
+                expected: "Unit",
+                found: "Foo".to_string()
+            })
+        );
+    }
+
+    #[test]
+    fn deserialize_newtype_struct() {
+        #[derive(Debug, PartialEq, Deserialize)]
+        struct Foo(i32);
+
+        let mut de = Decoder::new(b"i1995e");
+        assert_eq!(Foo::deserialize(&mut de), Ok(Foo(1995)));
+    }
+
+    #[test]
+    fn deserialize_map_ok() {
+        let mut map = HashMap::new();
+        map.insert("foo", "bar");
+
+        let mut de = Decoder::new(b"d3:foo3:bare");
+        assert_eq!(HashMap::deserialize(&mut de), Ok(map))
+    }
+
+    #[test]
+    fn deserialize_map_err() {
+        let mut de = Decoder::new(b"di1995e3:fooe");
+        assert_eq!(
+            HashMap::<i32, &str>::deserialize(&mut de),
+            Err(Error::Malformed)
+        )
+    }
 }
